"""
Distutils doesn't support nasm, so this is a custom compiler for NASM
"""
from distutils.errors import DistutilsExecError, CompileError

from distutils.msvc9compiler import MSVCCompiler
import os


class WinNasmCompiler(MSVCCompiler):
    compiler_type = 'winnasm'
    src_extensions = ['.asm']

    def __init__ (self,
                  verbose=0,
                  dry_run=0,
                  force=0):

        MSVCCompiler.__init__(self, verbose, dry_run, force)

    def initialize(self, plat_name=None):
        self.cc = self.find_exe("nasm.exe")
        self.linker = self.find_exe("link.exe")
        self.lib = self.find_exe("lib.exe")
        self.rc = self.find_exe("rc.exe")   # resource compiler
        self.mc = self.find_exe("mc.exe")   # message compiler

        self.compile_options = ["-f win64", "-DWINDOWS", "-DNOPIE"]
        self.compile_options_debug = ["-f win64", "-g", "-DWINDOWS", "-DNOPIE"]
        self.ldflags_shared = ['/DLL', '/nologo', '/INCREMENTAL:NO', '/NOENTRY']
        self.ldflags_shared_debug = ['/DLL', '/nologo', '/INCREMENTAL:NO', '/DEBUG', '/NOENTRY']
        self.ldflags_static = ['/nologo']
        self.initialized = True

    def link(self, target_desc, objects,
             output_filename, output_dir=None, libraries=None,
             library_dirs=None, runtime_library_dirs=None,
             export_symbols=None, debug=0, extra_preargs=None,
             extra_postargs=None, build_temp=None, target_lang=None):

        return super().link(target_desc, objects,
                            output_filename, output_dir, libraries,
                            library_dirs, runtime_library_dirs,
                            export_symbols, debug, extra_preargs,
                            extra_postargs, build_temp, target_lang)

    def compile(self, sources,
                output_dir=None, macros=None, include_dirs=None, debug=0,
                extra_preargs=None, extra_postargs=None, depends=None):

        if not self.initialized:
            self.initialize()
        compile_info = self._setup_compile(output_dir, macros, include_dirs,
                                           sources, depends, extra_postargs)
        macros, objects, extra_postargs, pp_opts, build = compile_info

        compile_opts = extra_preargs or []

        if debug:
            compile_opts.extend(self.compile_options_debug)
        else:
            compile_opts.extend(self.compile_options)

        for obj in objects:
            try:
                src, ext = build[obj]
            except KeyError:
                continue
            if debug:
                # pass the full pathname to MSVC in debug mode,
                # this allows the debugger to find the source file
                # without asking the user to browse for it
                src = os.path.abspath(src)

            input_opt =  src
            output_opt = "-o" + obj
            try:
                self.spawn([self.cc] + compile_opts + pp_opts +
                           [input_opt, output_opt] +
                           extra_postargs)
            except DistutilsExecError as msg:
                raise CompileError(msg)

        return objects



"""
Cython wrapper for the GzipReader and ParallelGzipReader C++ classes.
"""

from libc.stdlib cimport malloc, free
from libc.stdio cimport SEEK_SET
from libcpp.string cimport string
from libcpp.map cimport map
from libcpp.vector cimport vector
from libcpp cimport bool
from cpython.buffer cimport PyObject_GetBuffer, PyBuffer_Release, PyBUF_ANY_CONTIGUOUS, PyBUF_SIMPLE
from cpython.ref cimport PyObject

import builtins
import io
import os
import sys

ctypedef (unsigned long long int) size_t
ctypedef (long long int) lli

cdef extern from "tools/rapidgzip.cpp":
    int rapidgzipCLI(int, char**) except +

cdef extern from "rapidgzip/ParallelGzipReader.hpp" namespace "rapidgzip":
    cppclass ChunkData

    cppclass ParallelGzipReader[ChunkData, ENABLE_STATISTICS=*]:
        ParallelGzipReader(string, size_t) except +
        ParallelGzipReader(int, size_t) except +
        ParallelGzipReader(PyObject*, size_t) except +

        bool eof() except +
        int fileno() except +
        bool seekable() except +
        void close() except +
        bool closed() except +
        size_t seek(lli, int) except +
        size_t tell() except +
        size_t size() except +

        size_t tellCompressed() except +
        int read(int, char*, size_t) except +
        bool blockOffsetsComplete() except +
        map[size_t, size_t] blockOffsets() except +
        map[size_t, size_t] availableBlockOffsets() except +
        void importIndex(PyObject*) except +
        void exportIndex(PyObject*) except +
        void joinThreads() except +

    # To be used as template argument for ParallelGzipReader to avoid triggering the Cython
    # error message about template arguments being of an unknown type (because it is a value).
    # Must be inside an "extern" section or else Cython will try to generate a struct named
    # "true", which does not compile.
    cdef cppclass TrueValue "true":
        pass
    cdef cppclass RapidgzipChunkData "rapidgzip::ChunkData":
        pass

ctypedef ParallelGzipReader[RapidgzipChunkData, TrueValue] ParallelGzipReaderVerbose
ctypedef ParallelGzipReader[RapidgzipChunkData] ParallelGzipReaderNonVerbose


def _isFileObject(file):
    return (
        hasattr(file, 'read') and callable(getattr(file, 'read'))
        and hasattr(file, 'seek') and callable(getattr(file, 'seek'))
        and hasattr(file, 'tell') and callable(getattr(file, 'tell'))
        and hasattr(file, 'seekable') and callable(getattr(file, 'seekable'))
    )

def _hasValidFileno(file):
    if not hasattr(file, 'fileno'):
        return False

    try:
        fileno = file.fileno()
        return isinstance(fileno, int) and fileno >= 0
    except Exception:
        return False

cdef class _RapidgzipFile():
    cdef ParallelGzipReaderNonVerbose* gzipReader
    cdef ParallelGzipReaderVerbose* gzipReaderVerbose

    def __cinit__(self, file, parallelization, verbose = False):
        """
        file : can be a file path, a file descriptor, or a file object
               with suitable read, seekable, seek, tell methods.
        """
        # This should be done before any error handling because we cannot initialize members in-place in Cython!
        # nullptr exists but does not work: https://github.com/cython/cython/issues/3314
        self.gzipReader = NULL
        self.gzipReaderVerbose = NULL

        if not isinstance(parallelization, int):
            raise TypeError(f"Parallelization argument must be an integer not '{parallelization}'!")

        if verbose:
            if isinstance(file, int):
                self.gzipReaderVerbose = new ParallelGzipReaderVerbose(<int>file, <int>parallelization)
            elif _hasValidFileno(file):
                self.gzipReaderVerbose = new ParallelGzipReaderVerbose(<int>file.fileno(), <int>parallelization)
            elif _isFileObject(file):
                self.gzipReaderVerbose = new ParallelGzipReaderVerbose(<PyObject*>file, <int>parallelization)
            elif isinstance(file, basestring) and hasattr(file, 'encode'):
                # Note that BytesIO also is an instance of basestring but fortunately has no encode method!
                self.gzipReaderVerbose = new ParallelGzipReaderVerbose(<string>file.encode(), <int>parallelization)
        else:
            if isinstance(file, int):
                self.gzipReader = new ParallelGzipReaderNonVerbose(<int>file, <int>parallelization)
            elif _hasValidFileno(file):
                self.gzipReader = new ParallelGzipReaderNonVerbose(<int>file.fileno(), <int>parallelization)
            elif _isFileObject(file):
                self.gzipReader = new ParallelGzipReaderNonVerbose(<PyObject*>file, <int>parallelization)
            elif isinstance(file, basestring) and hasattr(file, 'encode'):
                # Note that BytesIO also is an instance of basestring but fortunately has no encode method!
                self.gzipReader = new ParallelGzipReaderNonVerbose(<string>file.encode(), <int>parallelization)

        if self.gzipReader == NULL and self.gzipReaderVerbose == NULL:
            raise Exception("Expected file name string, file descriptor integer, "
                            "or file-like object for ParallelGzipReader!")


    def __init__(self, *args, **kwargs):
        pass

    def __dealloc__(self):
        self.close()

        if self.gzipReader:
            del self.gzipReader
            self.gzipReader = NULL

        if self.gzipReaderVerbose:
            del self.gzipReaderVerbose
            self.gzipReaderVerbose = NULL

    def close(self):
        if self.gzipReaderVerbose != NULL and not self.gzipReaderVerbose.closed():
            self.gzipReaderVerbose.close()
        if self.gzipReader != NULL and not self.gzipReader.closed():
            self.gzipReader.close()

    def closed(self):
        return (
            ( self.gzipReader == NULL or self.gzipReader.closed() )
            and ( self.gzipReaderVerbose == NULL or self.gzipReaderVerbose.closed() )
        )

    def fileno(self):
        if self.gzipReader:
            return self.gzipReader.fileno()
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.fileno()
        raise Exception("Invalid file object!")

    def seekable(self):
        return (
            ( self.gzipReader != NULL and self.gzipReader.seekable() )
            or ( self.gzipReaderVerbose != NULL and self.gzipReaderVerbose.seekable() )
        )

    def readinto(self, bytes_like):
        if not self.gzipReader and not self.gzipReaderVerbose:
            raise Exception("Invalid file object!")

        bytes_count = 0

        cdef Py_buffer buffer
        PyObject_GetBuffer(bytes_like, &buffer, PyBUF_SIMPLE | PyBUF_ANY_CONTIGUOUS)
        try:
            if self.gzipReader:
                bytes_count = self.gzipReader.read(-1, <char*>buffer.buf, len(bytes_like))
            else:
                bytes_count = self.gzipReaderVerbose.read(-1, <char*>buffer.buf, len(bytes_like))
        finally:
            PyBuffer_Release(&buffer)

        return bytes_count

    def seek(self, offset, whence = io.SEEK_SET):
        if self.gzipReader:
            return self.gzipReader.seek(offset, whence)
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.seek(offset, whence)
        raise Exception("Invalid file object!")

    def tell(self):
        if self.gzipReader:
            return self.gzipReader.tell()
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.tell()
        raise Exception("Invalid file object!")

    def size(self):
        if self.gzipReader:
            return self.gzipReader.size()
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.size()
        raise Exception("Invalid file object!")

    def tell_compressed(self):
        if self.gzipReader:
            return self.gzipReader.tellCompressed()
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.tellCompressed()
        raise Exception("Invalid file object!")

    def block_offsets_complete(self):
        if self.gzipReader:
            return self.gzipReader.blockOffsetsComplete()
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.blockOffsetsComplete()
        raise Exception("Invalid file object!")

    def block_offsets(self):
        if self.gzipReader:
            return self.gzipReader.blockOffsets()
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.blockOffsets()
        raise Exception("Invalid file object!")

    def available_block_offsets(self):
        if self.gzipReader:
            return self.gzipReader.availableBlockOffsets()
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.availableBlockOffsets()
        raise Exception("Invalid file object!")

    def import_index(self, file):
        if self.gzipReader:
            if isinstance(file, str):
                with builtins.open(file, "rb") as fileObject:
                    return self.gzipReader.importIndex(<PyObject*>fileObject)
            return self.gzipReader.importIndex(<PyObject*>file)

        if self.gzipReaderVerbose:
            if isinstance(file, str):
                with builtins.open(file, "rb") as fileObject:
                    return self.gzipReaderVerbose.importIndex(<PyObject*>fileObject)
            return self.gzipReaderVerbose.importIndex(<PyObject*>file)

        raise Exception("Invalid file object!")

    def export_index(self, file):
        if self.gzipReader:
            if isinstance(file, str):
                with builtins.open(file, "wb") as fileObject:
                    return self.gzipReader.exportIndex(<PyObject*>fileObject)
            return self.gzipReader.exportIndex(<PyObject*>file)

        if self.gzipReaderVerbose:
            if isinstance(file, str):
                with builtins.open(file, "wb") as fileObject:
                    return self.gzipReaderVerbose.exportIndex(<PyObject*>fileObject)
            return self.gzipReaderVerbose.exportIndex(<PyObject*>file)

        raise Exception("Invalid file object!")

    def join_threads(self):
        if self.gzipReader:
            return self.gzipReader.joinThreads()
        if self.gzipReaderVerbose:
            return self.gzipReaderVerbose.joinThreads()


# Extra class because cdefs are not visible from outside but cdef class can't inherit from io.BufferedIOBase
# ParallelGzipReader has its own internal buffer. Using io.BufferedReader is not necessary and might even
# worsen performance for access from parallel processes because it will always try to rebuffer the new seemingly
# random access while the internal cache is able to keep multiple buffers for different offsets.
# Using io.BufferedReader degraded performance by almost 2x for the test case of calculating the CRC32 four times
# using four parallel find | xarg crc32 instances for a file containing 10k files with each 1 MiB of base64 data.
class RapidgzipFile(io.RawIOBase):
    def __init__(self, filename, parallelization = 0, verbose = False):
        self.gzipReader = _RapidgzipFile(filename, parallelization, verbose)
        self.name = filename if isinstance(filename, str) else ""
        self.mode = 'rb'

        self.readinto = self.gzipReader.readinto
        self.seek     = self.gzipReader.seek
        self.tell     = self.gzipReader.tell
        self.seekable = self.gzipReader.seekable

        self.tell_compressed         = self.gzipReader.tell_compressed
        self.block_offsets           = self.gzipReader.block_offsets
        self.export_index            = self.gzipReader.export_index
        self.import_index            = self.gzipReader.import_index
        self.block_offsets_complete  = self.gzipReader.block_offsets_complete
        self.available_block_offsets = self.gzipReader.available_block_offsets
        self.size                    = self.gzipReader.size

        if hasattr(self.gzipReader, 'join_threads'):
            self.join_threads = self.gzipReader.join_threads

        # IOBase provides sane default implementations for read, readline, readlines, readall, ...

    def fileno(self):
        try:
            return self.gzipReader.fileno()
        except Exception as exception:
            raise io.UnsupportedOperation() from exception

    def close(self):
        if self.closed:
            return
        super().close()
        self.gzipReader.close()

    def readable(self):
        return True


def open(filename, parallelization = 0, verbose = False):
    """
    filename: can be a file path, a file descriptor, or a file object
              with suitable read, seekable, seek, and tell methods.
    """
    return RapidgzipFile(filename, parallelization, verbose)


def cli():
    args = sys.argv
    cdef char** cargs = <char**> malloc(len(args) * sizeof(char*))
    cdef vector[Py_buffer] buffers
    buffers.resize(len(args))

    try:
        for i, arg in enumerate(args):
            PyObject_GetBuffer(arg.encode(), &buffers[i], PyBUF_SIMPLE | PyBUF_ANY_CONTIGUOUS)
            cargs[i] = <char*>buffers[i].buf

        return rapidgzipCLI(len(args), cargs)
    finally:
        free(cargs)
        for buffer in buffers:
            PyBuffer_Release(&buffer)


__version__ = '0.10.4'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import os
import platform
import shutil
import sys
import tempfile
from distutils.errors import CompileError

from setuptools import setup
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext

# This fallback is only for jinja, which is used by conda to analyze this setup.py before any build environment
# is set up.
try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

# ISA-l does not compile on 32-bit becaue it contains statements such as [bits 64].
# It simply is not supported and I also don't see a reason. 32-bit should be long dead exactly
# like almost all (96% according to Steam) PCs have AVX support.
# On aarch64 macOS, ISA-L causes: ImportError: dlopen(python3.10/site-packages/rapidgzip.cpython-310-darwin.so, 0x0002):
#   symbol not found in flat namespace '_decode_huffman_code_block_stateless'
# On aarch64 linux, ISA-L causes:
#   /bin/ld: external/isa-l/igzip/igzip_decode_block_stateless_01.obj: error adding symbols: file in wrong format
# -> I need to migrate the ARM build settings to CMake. ISA-L has aarch64 subfolders with assembly files.
withIsal = shutil.which("nasm") is not None and platform.machine() in ['x86_64', 'AMD64']

zlib_sources = ['inflate.c', 'crc32.c', 'adler32.c', 'inftrees.c', 'inffast.c', 'zutil.c']
zlib_sources = ['external/zlib/' + source for source in zlib_sources]

isal_sources = [
    # "include/igzip_lib.h",
    # "include/unaligned.h",
    "include/reg_sizes.asm",
    "include/multibinary.asm",
    "igzip/igzip_inflate.c",
    "igzip/igzip.c",
    "igzip/hufftables_c.c",
    # "igzip/igzip_checksums.h",
    "igzip/igzip_inflate_multibinary.asm",
    "igzip/igzip_decode_block_stateless_01.asm",
    "igzip/igzip_decode_block_stateless_04.asm",
    "igzip/rfc1951_lookup.asm",
    # "igzip/igzip_wrapper.h",
    # "igzip/static_inflate.h",
    "igzip/stdmac.asm",
]
isal_sources = ['external/isa-l/' + source for source in isal_sources] if withIsal else []

include_dirs = ['.', 'core', 'rapidgzip', 'rapidgzip/huffman', 'external/zlib', 'external/cxxopts/include',
                'external/rpmalloc/rpmalloc']
isal_includes = ['external/isa-l/include', 'external/isa-l/igzip']
if withIsal:
    include_dirs += isal_includes

extensions = [
    Extension(
        # fmt: off
        name         = 'rapidgzip',
        sources      = ['rapidgzip.pyx'] + zlib_sources + isal_sources + ['external/rpmalloc/rpmalloc/rpmalloc.c'],
        include_dirs = include_dirs,
        language     = 'c++',
        # fmt: on
    ),
]

if cythonize:
    extensions = cythonize(extensions, compiler_directives={'language_level': '3'})


def supportsFlag(compiler, flag):
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as file:
        file.write('int main() { return 0; }')
        try:
            compiler.compile([file.name], extra_postargs=[flag])
        except CompileError:
            print("[Info] Compiling with argument failed. Will try another one. The above error can be ignored!")
            return False
    return True


def hasInclude(compiler, systemInclude):
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as file:
        file.write(f'#include <{systemInclude}>\n' + 'int main() { return 0; }')
        try:
            compiler.compile([file.name])
        except CompileError:
            print(
                f"[Info] Check for {systemInclude} system header failed. Will try without out it. "
                "The above error can be ignored!"
            )
            return False
    return True


# https://github.com/cython/cython/blob/master/docs/src/tutorial/appendix.rst#python-38
class Build(build_ext):
    def build_extensions(self):
        # This is as hacky as it gets just in order to have different compile arguments for the zlib C-code as
        # opposed to the C++ code but I don't see another way with this subpar "build system" if you can call
        # it even that.
        oldCompile = self.compiler.compile

        if not withIsal:
            nasmCompiler = None
        elif sys.platform == "win32":
            from nasm_extension.winnasmcompiler import WinNasmCompiler

            nasmCompiler = WinNasmCompiler(verbose=True)
        else:
            from nasm_extension.nasmcompiler import NasmCompiler

            nasmCompiler = NasmCompiler(verbose=True)

        def newCompile(sources, *args, **kwargs):
            cSources = [source for source in sources if source.endswith('.c')]
            asmSources = [source for source in sources if source.endswith('.asm')]
            cppSources = [source for source in sources if not source.endswith('.c') and not source.endswith('.asm')]

            objects = []
            if asmSources and nasmCompiler:
                nasm_kwargs = copy.deepcopy(kwargs)
                nasm_kwargs['extra_postargs'] = []
                nasm_kwargs['include_dirs'] = isal_includes

                # One of the crudest hacks ever. But for some reason, I get:
                # fatal: unable to open include file `reg_sizes.asm'
                # But this only happens with the manylinux2014 image. Building with cibuildwheel and manylinux2_28
                # works perfectly fine without this hack. Maybe it is a problem with 2.10.07-7.el7 (manylinux2014),
                # which has been fixed in 2.15.03-3.el8 (manylinux_2_28).
                # @see https://www.nasm.us/xdoc/2.16.01/html/nasmdocc.html#section-C.1.32
                # The only mention of -I, which I could find, is in 2.14:
                # > Changed -I option semantics by adding a trailing path separator unconditionally.
                # It even fails to include .asm files in the same directory as the .asm file to be compiled.
                #   nasm -I. -Iexternal/isa-l/include -Iexternal/isa-l/igzip -f elf64 \
                #       external/isa-l/igzip/igzip_decode_block_stateless_01.asm -o \
                #       build/temp.linux-x86_64-3.6/external/isa-l/igzip/igzip_decode_block_stateless_01.obj
                # external/isa-l/igzip/igzip_decode_block_stateless_01.asm:3: fatal: unable to open include file
                #   `igzip_decode_block_stateless.asm'
                # error: command 'nasm' failed with exit status 1
                # I even tried chaning -I<dir> to -I <dir> by overwriting nasmcompiler._setup_compile but to no avail.
                for path in isal_includes:
                    for fileName in os.listdir(path):
                        if fileName.endswith('.asm'):
                            shutil.copy(os.path.join(path, fileName), ".")
                objects.extend(nasmCompiler.compile(asmSources, *args, **nasm_kwargs))

            if cppSources:
                objects.extend(oldCompile(cppSources, *args, **kwargs))

            if cSources:
                cppCompileArgs = [
                    '-fconstexpr-ops-limit=99000100',
                    '-fconstexpr-steps=99000100',
                    '-std=c++17',
                    '/std:c++17',
                ]
                if 'extra_postargs' in kwargs:
                    kwargs['extra_postargs'] = [x for x in kwargs['extra_postargs'] if x not in cppCompileArgs]
                objects.extend(oldCompile(cSources, *args, **kwargs))

            return objects

        self.compiler.compile = newCompile

        for ext in self.extensions:
            ext.extra_compile_args = [
                '-std=c++17',
                '-O3',
                '-DNDEBUG',
                '-DWITH_PYTHON_SUPPORT',
                '-DWITH_RPMALLOC',
                '-D_LARGEFILE64_SOURCE=1',
                '-D_FORTIFY_SOURCE=2',
                '-D_GLIBCXX_ASSERTIONS',
            ]
            if nasmCompiler:
                ext.extra_compile_args.append('-DWITH_ISAL')

            # https://github.com/cython/cython/issues/2670#issuecomment-432212671
            # https://github.com/cython/cython/issues/3405#issuecomment-596975159
            # https://bugs.python.org/issue35037
            # https://bugs.python.org/issue4709
            if platform.system() == 'Windows' and platform.machine().endswith('64'):
                ext.extra_compile_args += ['-DMS_WIN64']

            if self.compiler.compiler_type == 'mingw32':
                ext.extra_link_args = [
                    '-static-libgcc',
                    '-static-libstdc++',
                    '-Wl,-Bstatic,--whole-archive',
                    '-lwinpthread',
                    '-Wl,--no-whole-archive',
                ]

            elif self.compiler.compiler_type == 'msvc':
                ext.extra_compile_args = [
                    '/std:c++17',
                    '/O2',
                    '/DNDEBUG',
                    '/DWITH_PYTHON_SUPPORT',
                    '/DWITH_RPMALLOC',
                    '/constexpr:steps99000100',
                ]
                if nasmCompiler:
                    ext.extra_compile_args.append('/DWITH_ISAL')
                # This list is from rpmalloc/build/ninja/msvc.py
                ext.libraries = ['kernel32', 'user32', 'shell32', 'advapi32']
            else:
                # The default limit is ~33 M (1<<25) and 99 M seem to be enough to compile currently on GCC 11.
                if supportsFlag(self.compiler, '-fconstexpr-ops-limit=99000100'):
                    ext.extra_compile_args += ['-fconstexpr-ops-limit=99000100']
                elif supportsFlag(self.compiler, '-fconstexpr-steps=99000100'):
                    ext.extra_compile_args += ['-fconstexpr-steps=99000100']

                if sys.platform.startswith('darwin') and supportsFlag(self.compiler, '-mmacosx-version-min=10.14'):
                    ext.extra_compile_args += ['-mmacosx-version-min=10.14']
                    ext.extra_link_args += ['-mmacosx-version-min=10.14']

                # Add some hardening. See e.g.:
                # https://www.phoronix.com/news/GCC-fhardened-Hardening-Option
                # https://developers.redhat.com/blog/2018/03/21/compiler-and-linker-flags-gcc
                # I have not observed any performance impact for these.
                ext.extra_compile_args += ['-fstack-protector-strong', '-fstack-clash-protection']
                # AppleClang seems to not like this flag:
                if supportsFlag(self.compiler, '-fcf-protection=full'):
                    ext.extra_compile_args += ['-fcf-protection=full']

            if hasInclude(self.compiler, 'unistd.h'):
                ext.extra_compile_args += ['-DZ_HAVE_UNISTD_H']

        super(Build, self).build_extensions()


setup(
    # fmt: off
    ext_modules = extensions,
    cmdclass    = {'build_ext': Build},
    # fmt: on
)
